@@ -1,4 +1,4 @@
-/*   
+/*
  * Copyright (C) 2003 Robert Kooima
  *
  * NEVERPUTT is  free software; you can redistribute  it and/or modify
@@ -19,18 +19,20 @@
 #include "game.h"
 #include "vec3.h"
 #include "geom.h"
+#include "ball.h"
 #include "back.h"
 #include "hole.h"
 #include "hud.h"
 #include "image.h"
 #include "audio.h"
-#include "solid.h"
+#include "solid_gl.h"
+#include "solid_phys.h"
 #include "config.h"
 
 /*---------------------------------------------------------------------------*/
 
 static struct s_file file;
-static short         ball;
+static int           ball;
 
 static float view_a;                    /* Ideal view rotation about Y axis  */
 static float view_m;
@@ -43,7 +45,6 @@ static float view_v[3];                 /* Current view vector               */
 static float view_p[3];                 /* Current view position             */
 static float view_e[3][3];              /* Current view orientation          */
 
-static int   swch_e = 1;                /* Switching enabled flag            */
 static float jump_e = 1;                /* Jumping enabled flag              */
 static float jump_b = 0;                /* Jump-in-progress flag             */
 static float jump_dt;                   /* Jump duration                     */
@@ -84,13 +85,13 @@ void game_init(const char *s)
     jump_b = 0;
 
     view_init();
-    sol_load(&file, config_data(s), config_get_d(CONFIG_TEXTURES),
+    sol_load_gl(&file, config_data(s), config_get_d(CONFIG_TEXTURES),
                                     config_get_d(CONFIG_SHADOW));
 }
 
 void game_free(void)
 {
-    sol_free(&file);
+    sol_free_gl(&file);
 }
 
 /*---------------------------------------------------------------------------*/
@@ -107,7 +108,7 @@ static void game_draw_vect_prim(const struct s_file *fp, GLenum mode)
     v_cpy(z, view_e[2]);
 
     r = fp->uv[ball].r;
-    
+
     glBegin(mode);
     {
         glColor4f(1.0f, 1.0f, 0.5f, 0.5f);
@@ -155,7 +156,8 @@ static void game_draw_vect(const struct s_file *fp)
     }
 }
 
-static void game_draw_balls(const struct s_file *fp)
+static void game_draw_balls(const struct s_file *fp,
+                            const float *bill_M, float t)
 {
     static const GLfloat color[5][4] = {
         { 1.0f, 1.0f, 1.0f, 0.7f },
@@ -165,28 +167,31 @@ static void game_draw_balls(const struct s_file *fp)
         { 1.0f, 1.0f, 0.0f, 1.0f },
     };
 
-    float M[16];
     int ui;
 
     for (ui = curr_party(); ui > 0; ui--)
     {
         if (ui == ball)
         {
+            float ball_M[16];
+            float pend_M[16];
+
+            m_basis(ball_M, fp->uv[ui].e[0], fp->uv[ui].e[1], fp->uv[ui].e[2]);
+            m_basis(pend_M, fp->uv[ui].E[0], fp->uv[ui].E[1], fp->uv[ui].E[2]);
+
             glPushMatrix();
             {
-                m_basis(M, fp->uv[ui].e[0], fp->uv[ui].e[1], fp->uv[ui].e[2]);
-
                 glTranslatef(fp->uv[ui].p[0],
                              fp->uv[ui].p[1] + BALL_FUDGE,
                              fp->uv[ui].p[2]);
-                glMultMatrixf(M);
                 glScalef(fp->uv[ui].r,
                          fp->uv[ui].r,
                          fp->uv[ui].r);
 
+                glEnable(GL_COLOR_MATERIAL);
                 glColor4fv(color[ui]);
-
-                ball_draw();
+                ball_draw(ball_M, pend_M, bill_M, t);
+                glDisable(GL_COLOR_MATERIAL);
             }
             glPopMatrix();
         }
@@ -210,9 +215,10 @@ static void game_draw_balls(const struct s_file *fp)
             glPopMatrix();
         }
     }
+    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
 }
 
-static void game_draw_goals(const struct s_file *fp, float rx, float ry)
+static void game_draw_goals(const struct s_file *fp)
 {
     int zi;
 
@@ -242,7 +248,7 @@ static void game_draw_jumps(const struct s_file *fp)
                          fp->jv[ji].p[2]);
 
             glScalef(fp->jv[ji].r, 1.f, fp->jv[ji].r);
-            jump_draw();
+            jump_draw(!jump_e);
         }
         glPopMatrix();
     }
@@ -254,6 +260,9 @@ static void game_draw_swchs(const struct s_file *fp)
 
     for (xi = 0; xi < fp->xc; xi++)
     {
+        if (fp->xv[xi].i)
+            continue;
+
         glPushMatrix();
         {
             glTranslatef(fp->xv[xi].p[0],
@@ -261,7 +270,7 @@ static void game_draw_swchs(const struct s_file *fp)
                          fp->xv[xi].p[2]);
 
             glScalef(fp->xv[xi].r, 1.f, fp->xv[xi].r);
-            swch_draw(fp->xv[xi].f);
+            swch_draw(fp->xv[xi].f, fp->xv[xi].e);
         }
         glPopMatrix();
     }
@@ -269,12 +278,17 @@ static void game_draw_swchs(const struct s_file *fp)
 
 /*---------------------------------------------------------------------------*/
 
-void game_draw(int pose)
+void game_draw(int pose, float t)
 {
+    static const float a[4] = { 0.2f, 0.2f, 0.2f, 1.0f };
+    static const float s[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
+    static const float e[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
+    static const float h[1] = { 0.0f };
+
     const float light_p[4] = { 8.f, 32.f, 8.f, 1.f };
 
     const struct s_file *fp = &file;
-    
+
     float fov = FOV;
 
     if (jump_b) fov *= 2.0f * fabsf(jump_dt - 0.5f);
@@ -283,7 +297,10 @@ void game_draw(int pose)
     glPushAttrib(GL_LIGHTING_BIT);
     glPushMatrix();
     {
-        float v[3], rx, ry;
+        float T[16], M[16], v[3], rx, ry;
+
+        m_view(T, view_c, view_p, view_e[1]);
+        m_xps(M, T);
 
         v_sub(v, view_c, view_p);
 
@@ -291,8 +308,7 @@ void game_draw(int pose)
         ry = V_DEG(fatan2f(+v[0], -v[2]));
 
         glTranslatef(0.f, 0.f, -v_len(v));
-        glRotatef(rx, 1.f, 0.f, 0.f);
-        glRotatef(ry, 0.f, 1.f, 0.f);
+        glMultMatrixf(M);
         glTranslatef(-view_c[0], -view_c[1], -view_c[2]);
 
         /* Center the skybox about the position of the camera. */
@@ -309,7 +325,7 @@ void game_draw(int pose)
 
         /* Draw the floor. */
 
-        sol_draw(fp);
+        sol_draw(fp, 0, 1);
 
         if (config_get_d(CONFIG_SHADOW) && !pose)
         {
@@ -325,13 +341,29 @@ void game_draw(int pose)
 
         if (pose == 0)
         {
-            game_draw_balls(fp);
+            game_draw_balls(fp, T, t);
             game_draw_vect(fp);
         }
 
-        game_draw_goals(fp, -rx, -ry);
-        game_draw_jumps(fp);
-        game_draw_swchs(fp);
+        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,   a);
+        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,  s);
+        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION,  e);
+        glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, h);
+
+        game_draw_goals(fp);
+
+        glEnable(GL_COLOR_MATERIAL);
+        glDisable(GL_LIGHTING);
+        glDisable(GL_TEXTURE_2D);
+        glDepthMask(GL_FALSE);
+        {
+            game_draw_jumps(fp);
+            game_draw_swchs(fp);
+        }
+        glDepthMask(GL_TRUE);
+        glEnable(GL_TEXTURE_2D);
+        glEnable(GL_LIGHTING);
+        glDisable(GL_COLOR_MATERIAL);
     }
     glPopMatrix();
     glPopAttrib();
@@ -413,7 +445,6 @@ static int game_update_state(float dt)
 
     struct s_file *fp = &file;
     float p[3];
-    int e = swch_e;
 
     if (dt > 0.f)
         t += dt;
@@ -422,7 +453,7 @@ static int game_update_state(float dt)
 
     /* Test for a switch. */
 
-    if ((swch_e = sol_swch_test(fp, swch_e, ball)) != e && e)
+    if (sol_swch_test(fp, ball))
         audio_play(AUD_SWITCH, 1.f);
 
     /* Test for a jump. */
@@ -432,7 +463,7 @@ static int game_update_state(float dt)
         jump_b  = 1;
         jump_e  = 0;
         jump_dt = 0.f;
-        
+
         audio_play(AUD_JUMP, 1.f);
     }
     if (jump_e == 0 && jump_b == 0 &&  sol_jump_test(fp, jump_p, ball) == 0)
@@ -469,7 +500,7 @@ static int game_update_state(float dt)
  * the measured  frame time  exceeds this  maximum, we cut  the time  step in
  * half, and  do two updates.  If THIS  time step exceeds the  maximum, we do
  * four updates.  And  so on.  In this way, the physics  system is allowed to
- * seek an optimal update rate independant of, yet in integral sync with, the
+ * seek an optimal update rate independent of, yet in integral sync with, the
  * graphics frame rate.
  */
 
@@ -516,7 +547,7 @@ int game_step(const float g[3], float dt)
         for (i = 0; i < n; i++)
         {
             d = sol_step(fp, g, t, ball, &m);
-            
+
             if (b < d)
                 b = d;
             if (m)
